package csci1902.hw3;

import java.awt.Point;
import java.util.*;

/**
 * Test cases for common ant functionality
 * @author Michael Ekstrand <ekstrand@cs.umn.edu>
 *
 */
public class ForagerAntTest extends AntTestBase {
    private static final double DELTA = 0.0001;
    private static final int MAX_FOOD = 20;
    private ForagerAnt ant;
    private List<Entity> neighbors;
    
    protected void setUp() {
        super.setUp();
        ant = new ForagerAnt(MAX_FOOD, world, new Point(5,7));;
        ant.setPosition(new Point(1,1));
        neighbors = new ArrayList<Entity>();
    }
    
    protected Ant getAnt() {
        return ant;
    }
    
    /**
     * Make sure that the ant's initial state is correct (position, home
     * position, food).
     */
    public void testInitialState() {
        assertSame(world, ant.getWorld());
        assertEquals(0, ant.getFoodUnits());
        assertEquals(new Point(5,7), ant.getHomePosition());
        assertEquals(new Point(1,1), ant.getPosition());
    }
    
    /**
     * Make sure that the ant moves randomly correctly.  This test method
     * depends on knowledge of how the moveRandomly() method works, and preseeds
     * the fake RNG to get a predictable movement.
     */
    public void testRandomMovement() {
        rng.addNumber(0.05); // northwest motion
        ant.act(neighbors);
        assertEquals(new Point(1,1), ant.getPosition());
        world.commit();
        assertEquals(0, world.getFoodPheromone(1,1), DELTA);
        assertTrue(neighbors.isEmpty());
        
        ant.update();
        assertEquals(0, world.getFoodPheromone(1,1), DELTA);
        assertEquals(new Point(0,0), ant.getPosition());
        assertEquals(0, world.getFoodPheromone(0,0), DELTA);
        rng.addNumber(0.05);  // east motion in this case - will be the first
        ant.act(neighbors);
        ant.update();
        assertEquals(new Point(1, 0), ant.getPosition());
        assertEquals(0, world.getFoodPheromone(1,0), DELTA);
    }
    
    /**
     * Test the pheremone-tracking code.
     */
    public void testPheremoneTrack() {
        world.depositFoodPheromone(2, 1, 1);
        world.depositFoodPheromone(2, 0, 2);
        world.depositFoodPheromone(0, 0, 2);
        world.depositFoodPheromone(1, 2, 3);
        world.commit();
        rng.setDefault(0.7);
        ant.act(neighbors);
        ant.update();
        assertEquals(new Point(2, 1), ant.getPosition());
    }
    
    /**
     * Test that an ant collects food.
     */
    public void testCollectFood() {
        FoodSource fs = new FoodSource(world, new Point(2,2), 100);
        neighbors.add(fs);
        ant.act(neighbors);
        assertEquals(0, ant.getFoodUnits());
        ant.update();
        fs.update();
        assertEquals(MAX_FOOD, ant.getFoodUnits());
        assertEquals(100 - MAX_FOOD, fs.getFoodUnits());
        assertEquals(new Point(1,1), ant.getPosition());
    }
    
    /**
     * Gives the ant some food for interactions with food sinks, etc.
     * @param amount The max amount to feed the ant.
     */
    private void feedAnt(int amount) {
        FoodSource fs = new FoodSource(world, new Point(2,2), amount);
        List<Entity> neighbors = new ArrayList<Entity>();
        neighbors.add(fs);
        ant.act(neighbors);
        ant.update();
        fs.update();
    }
    
    /**
     * Test that an ant deposits food.
     */
    public void testDepositFood() {
        feedAnt(100);
        
        Colony c = new Colony(world, new Point(0, 1));
        neighbors.add(c);
        ant.act(neighbors);
        assertEquals(MAX_FOOD, ant.getFoodUnits());
        ant.update();
        c.update();
        assertEquals(0, ant.getFoodUnits());
        assertEquals(MAX_FOOD, c.getFoodLevel());
        assertEquals(new Point(1,1), ant.getPosition());
    }
    
    /**
     * Test that an ant deposits pheromone as it moves while carrying food.
     */
    public void testDepositPheromone() {
        feedAnt(100);
        
        ant.act(neighbors);
        assertEquals(new Point(1,1), ant.getPosition());
        ant.update();
        world.commit();
        // make sure that the ant deposits pheromone where it was
        assertEquals(5, world.getFoodPheromone(1, 1), DELTA);
        assertEquals(0, world.getFoodPheromone(2, 2), DELTA);
    }
    
    /**
     * Test that an ant goes home when carrying food.
     */
    public void testGoHome() {
        feedAnt(100);
        ant.act(neighbors);
        assertEquals(new Point(1,1), ant.getPosition());
        ant.update();
        assertEquals(new Point(2, 2), ant.getPosition());
        ant.act(neighbors);
        ant.update();
        assertEquals(new Point(3,3), ant.getPosition());
        ant.act(neighbors);
        ant.update();
        assertEquals(new Point(3,4), ant.getPosition());
        ant.act(neighbors);
        ant.update();
        assertEquals(new Point(4,5), ant.getPosition());
        ant.act(neighbors);
        ant.update();
        assertEquals(new Point(4,6), ant.getPosition());
        ant.act(neighbors);
        ant.update();
        assertEquals(new Point(5,7), ant.getPosition());
    }
    
    /**
     * Test that it prefers to collect food even when a food sink is around.
     */
    public void testCollectWithSink() {
        FoodSource source = new FoodSource(world, new Point(0,2), 1000);
        neighbors.add(source);
        neighbors.add(world.getColony());
        ant.act(neighbors);
        source.update();
        ant.update();
        assertEquals(MAX_FOOD, ant.getFoodUnits());
        assertEquals(0, world.getColony().getFoodLevel());
        assertEquals(1000 - MAX_FOOD, source.getFoodUnits());
    }
    
    /**
     * Test that it drops food even if there's a source.
     */
    public void testDepositWithSource() {
        feedAnt(100);
        
        FoodSource source = new FoodSource(world, new Point(0,2), 1000);
        neighbors.add(source);
        neighbors.add(world.getColony());
        ant.act(neighbors);
        source.update();
        ant.update();
        world.getColony().update();
        assertEquals(0, ant.getFoodUnits());
        assertEquals(MAX_FOOD, world.getColony().getFoodLevel());
        assertEquals(1000, source.getFoodUnits());
    }
    
    /**
     * Test that the ant skips a source if carrying food.
     */
    public void testSkipSourceWithFood() {
        feedAnt(5);
        
        FoodSource source = new FoodSource(world, new Point(0,2), 1000);
        neighbors.add(source);
        Point p = ant.getPosition();
        ant.act(neighbors);
        source.update();
        ant.update();
        assertEquals(5, ant.getFoodUnits());
        assertEquals(1000, source.getFoodUnits());
        assertFalse(p.equals(ant.getPosition()));
    }
    
    /**
     * Test that the ant only looks at one source.
     */
    public void testCollectFromOneSource() {
        FoodSource s1, s2, s3;
        s1 = new FoodSource(world, new Point(0,2), 0);
        s2 = new FoodSource(world, new Point(2,2), 10);
        s3 = new FoodSource(world, new Point(1,1), 1000);
        neighbors.add(s1);
        neighbors.add(s2);
        neighbors.add(s3);
        
        ant.act(neighbors);
        ant.update();
        s1.update();
        s2.update();
        s3.update();
        assertEquals(new Point(1,1), ant.getPosition());
        assertEquals(0, ant.getFoodUnits());
        assertEquals(10, s2.getFoodUnits());
        assertEquals(1000, s3.getFoodUnits());
    }
    
    /**
     * test that the ant skips a sink if it doesn't have food.
     */
    public void testSkipSinkWithoutFood() {
        neighbors.add(world.getColony());
        Point p = (Point) ant.getPosition().clone();
        ant.act(neighbors);
        ant.update();
        world.getColony().update();
        assertFalse(p.equals(ant.getPosition()));
        assertEquals(0, world.getColony().getFoodLevel());
    }
    
    /**
     * test that the ant correctly collects and deposits a parital food quantity.
     */
    public void testCollectDepositPartial() {
        feedAnt(5);
        assertEquals(5, ant.getFoodUnits());
        
        Colony c = new Colony(world, new Point(0,0));
        neighbors.add(c);
        ant.act(neighbors);
        c.update();
        ant.update();
        assertEquals(0, ant.getFoodUnits());
        assertEquals(5, c.getFoodLevel());
    }
    
    /**
     * Test that an ant only deposits in one food sink.
     */
    public void testDepositOneSink() {
        feedAnt(100);
        
        Colony c1 = new Colony(world, new Point(0,0));
        Colony c2 = new Colony(world, new Point(2,1));
        neighbors.add(c1);
        neighbors.add(c2);
        
        ant.act(neighbors);
        ant.update();
        c1.update();
        c2.update();
        
        assertEquals(0, ant.getFoodUnits());
        assertTrue((c1.getFoodLevel() == MAX_FOOD && c2.getFoodLevel() == 0)
                || (c1.getFoodLevel() == 0 && c2.getFoodLevel() == MAX_FOOD));
    }
    
    /**
     * Longer test of ant behavior when picking up food.
     */
    public void testTrackTowardsHome() {
        FoodSource fs1 = new FoodSource(world, new Point(1,0), 1000);
        neighbors.add(fs1);
        ant.act(neighbors);
        ant.update();
        fs1.update();
        world.commit();
        assertEquals(MAX_FOOD, ant.getFoodUnits());
        assertEquals(new Point(1,1), ant.getPosition());
        assertTrue(ant.isAlive());
        assertEquals(1000 - MAX_FOOD, fs1.getFoodUnits());
        assertEquals(0, world.getFoodPheromone(1, 1), DELTA);
        // make sure it didn't drop pheromone at the source
        assertEquals(0, world.getFoodPheromone(0, 1), DELTA);
        
        ant.act(neighbors);
        ant.update();
        fs1.update();
        world.commit();
        assertEquals(new Point(2,2), ant.getPosition());
        assertEquals(MAX_FOOD, ant.getFoodUnits());
        assertEquals(1000 - MAX_FOOD, fs1.getFoodUnits());
        assertTrue(ant.isAlive());
        assertEquals(5.0, world.getFoodPheromone(1,1), DELTA);
        
        neighbors.remove(fs1);
        ant.act(neighbors);
        ant.update();
        fs1.update();
        world.commit();
        assertEquals(new Point(3,3), ant.getPosition());
        assertEquals(MAX_FOOD, ant.getFoodUnits());
        assertEquals(1000 - MAX_FOOD, fs1.getFoodUnits());
        assertTrue(ant.isAlive());
        assertEquals(5.0, world.getFoodPheromone(2,2), DELTA);
    }
}
