package csci1902.hw3;

import java.awt.Point;
import java.util.*;

/**
 * Test cases for waste transport ant functionality
 * @author Michael Ekstrand <ekstrand@cs.umn.edu>
 *
 */
public class WasteTransportAntTest extends AntTestBase {
    private final static double DELTA = 0.0001;
    private final static int MAX_WASTE = 20;
    
    private WasteTransportAnt ant;
    private List<Entity> neighbors;
    
    /**
     * Construct a new waste transport ant to test its behavior.
     */
    protected void setUp() {
        super.setUp();
        ant = new WasteTransportAnt(MAX_WASTE, world, new Point(5,7));
        ant.setPosition(new Point(1,1));
        neighbors = new ArrayList<Entity>();
    }
    
    /**
     * retrieve the test ant
     */
    protected Ant getAnt() {
        return ant;
    }
    
    /**
     * Make sure that the ant's initial state is correct (position, home
     * position, food).
     */
    public void testInitialState() {
        assertSame(world, ant.getWorld());
        assertEquals(0, ant.getWasteUnits());
        assertEquals(new Point(5,7), ant.getHomePosition());
    }
    
    /**
     * Test that a waste ant goes home when empty.
     */
    public void testGoHome() {
        ant.act(neighbors);
        assertEquals(new Point(1,1), ant.getPosition());
        ant.update();
        assertEquals(new Point(2, 2), ant.getPosition());
        ant.act(neighbors);
        ant.update();
        assertEquals(new Point(3,3), ant.getPosition());
        ant.act(neighbors);
        ant.update();
        assertEquals(new Point(3,4), ant.getPosition());
        ant.act(neighbors);
        ant.update();
        assertEquals(new Point(4,5), ant.getPosition());
        ant.act(neighbors);
        ant.update();
        assertEquals(new Point(4,6), ant.getPosition());
        ant.act(neighbors);
        ant.update();
        assertEquals(new Point(5,7), ant.getPosition());
    }
    
    /**
     * Test that an ant deposits pheromone as it moves while empty.
     */
    public void testDepositPheromone() {
        ant.act(neighbors);
        assertEquals(new Point(1,1), ant.getPosition());
        ant.update();
        world.commit();
        // make sure that the ant deposits pheromone where it was
        assertEquals(5, world.getWastePheromone(1, 1), DELTA);
        assertEquals(0, world.getWastePheromone(2, 2), DELTA);
    }
    
    /**
     * Test that an ant collects waste.
     */
    public void testCollectWaste() {
        Colony c = new Colony(world, new Point(2,2));
        neighbors.add(c);
        ant.act(neighbors);
        assertEquals(0, ant.getWasteUnits());
        ant.update();
        c.update();
        assertEquals(MAX_WASTE, ant.getWasteUnits());
        assertEquals(new Point(1,1), ant.getPosition());
    }
    
    /**
     * Gives the ant some waste for interactions with waste sinks, etc.
     */
    private void giveAntWaste() {
        Colony colony = new Colony(world, new Point(2,2));
        List<Entity> neighbors = new ArrayList<Entity>();
        neighbors.add(colony);
        ant.act(neighbors);
        ant.update();
        colony.update();
    }
    
    /**
     * Test that an ant deposits waste.
     */
    public void testDepositWaste() {
        giveAntWaste();
        
        WasteDump dump = new WasteDump(world, new Point(0, 1));
        neighbors.add(dump);
        ant.act(neighbors);
        assertEquals(MAX_WASTE, ant.getWasteUnits());
        ant.update();
        dump.update();
        assertEquals(0, ant.getWasteUnits());
        assertEquals(MAX_WASTE, dump.getWasteUnits());
        assertEquals(new Point(1,1), ant.getPosition());
    }
    
    /**
     * Make sure that the ant moves randomly correctly.  This test method
     * depends on knowledge of how the moveRandomly() method works, and preseeds
     * the fake RNG to get a predictable movement.
     */
    public void testRandomMovement() {
        giveAntWaste();
        rng.addNumber(0.05); // northwest motion
        ant.act(neighbors);
        assertEquals(new Point(1,1), ant.getPosition());
        world.commit();
        assertEquals(0, world.getWastePheromone(1,1), DELTA);
        
        ant.update();
        assertEquals(0, world.getWastePheromone(1,1), DELTA);
        assertEquals(new Point(0,0), ant.getPosition());
        assertEquals(0, world.getWastePheromone(0,0), DELTA);
        rng.addNumber(.05);  // east motion in this case - will be the first
        ant.act(neighbors);
        ant.update();
        assertEquals(new Point(1, 0), ant.getPosition());
        assertEquals(0, world.getWastePheromone(1,0), DELTA);
    }
    
    /**
     * Test the pheremone-tracking code.
     */
    public void testPheremoneTrack() {
        giveAntWaste();
        world.depositWastePheromone(2, 1, 1);
        world.depositWastePheromone(2, 0, 2);
        world.depositWastePheromone(0, 0, 2);
        world.depositWastePheromone(1, 2, 3);
        world.commit();
        rng.setDefault(0.7);
        ant.act(neighbors);
        ant.update();
        assertEquals(new Point(2, 1), ant.getPosition());
    }
    
    /**
     * Test that it prefers to collect waste even when a waste sink is around.
     */
    public void testCollectWithSink() {
        Colony source = new Colony(world, new Point(0,2));
        neighbors.add(source);
        neighbors.add(world.getColony());
        ant.act(neighbors);
        source.update();
        ant.update();
        assertEquals(MAX_WASTE, ant.getWasteUnits());
        assertEquals(0, world.getColony().getFoodLevel());
    }
    
    /**
     * Test that it drops waste even if there's a source.
     */
    public void testDepositWithSource() {
        giveAntWaste();
        
        WasteDump dump = new WasteDump(world, new Point(0,0));
        neighbors.add(world.getColony());
        neighbors.add(dump);
        ant.act(neighbors);
        dump.update();
        ant.update();
        world.getColony().update();
        assertEquals(0, ant.getWasteUnits());
        assertEquals(MAX_WASTE, dump.getWasteUnits());
    }
    
    /**
     * test that the ant skips a sink if it doesn't have waste.
     */
    public void testSkipSinkWithoutWaste() {
        WasteDump dump = new WasteDump(world, new Point(0,0));
        neighbors.add(dump);
        Point p = (Point) ant.getPosition().clone();
        ant.act(neighbors);
        ant.update();
        world.getColony().update();
        assertFalse(p.equals(ant.getPosition()));
        assertEquals(0, dump.getWasteUnits());
    }
    
    public void testDepositOneSink() {
        giveAntWaste();
        
        WasteDump c1 = new WasteDump(world, new Point(0,0));
        WasteDump c2 = new WasteDump(world, new Point(2,1));
        neighbors.add(c1);
        neighbors.add(c2);
        
        ant.act(neighbors);
        ant.update();
        c1.update();
        c2.update();
        
        assertEquals(0, ant.getWasteUnits());
        assertTrue((c1.getWasteUnits() == MAX_WASTE && c2.getWasteUnits() == 0)
                || (c1.getWasteUnits() == 0 && c2.getWasteUnits() == MAX_WASTE));
    }
    
    /**
     * Longer test of ant behavior when dropping off waste.
     */
    public void testTrackTowardsHome() {
        giveAntWaste();
        
        WasteDump dump = new WasteDump(world, new Point(1,0));
        neighbors.add(dump);
        ant.act(neighbors);
        ant.update();
        dump.update();
        world.commit();
        assertEquals(0, ant.getWasteUnits());
        assertEquals(new Point(1,1), ant.getPosition());
        assertTrue(ant.isAlive());
        assertEquals(MAX_WASTE, dump.getWasteUnits());
        assertEquals(0, world.getWastePheromone(1, 1), DELTA);
        // make sure it didn't drop pheromone at the dump
        assertEquals(0, world.getWastePheromone(0, 1), DELTA);
        
        ant.act(neighbors);
        ant.update();
        dump.update();
        world.commit();
        assertEquals(new Point(2,2), ant.getPosition());
        assertEquals(0, ant.getWasteUnits());
        assertEquals(MAX_WASTE, dump.getWasteUnits());
        assertTrue(ant.isAlive());
        assertEquals(5.0, world.getWastePheromone(1,1), DELTA);
        
        neighbors.remove(dump);
        ant.act(neighbors);
        ant.update();
        dump.update();
        world.commit();
        assertEquals(new Point(3,3), ant.getPosition());
        assertEquals(0, ant.getWasteUnits());
        assertEquals(MAX_WASTE, dump.getWasteUnits());
        assertTrue(ant.isAlive());
        assertEquals(5.0, world.getWastePheromone(2,2), DELTA);
    }
}
